{*******************************************************}
{                                                       }
{            EWriter Help VCL Implementation            }
{               Based on eViewer.exe v3.0               }
{                 Released Dec 23, 2022                 }
{                                                       }
{          Copyright(c) 2022 EC Software GmbH           }
{                http://www.e-writer.com                }
{                                                       }
{*******************************************************}

unit Vcl.EwriterHelpViewer;

interface

uses System.Classes,
     ShellAPI,
     System.HelpIntfs, System.SysUtils, Winapi.Windows;

{$IF CompilerVersion >= 32}
  {$DEFINE CONTEXID_IS_THELPCONTEXT}
{$ENDIF}

{.$DEFINE EVIEWER250} //enable this compiler switch for eViewer < version 3.0

type
  TEWriterHelpViewer = class(TInterfacedObject, ICustomHelpViewer, IExtendedHelpViewer)
  private
    FViewerID: Integer;
    FInitialized: Boolean;
    FHelpManager: IHelpManager;
    FViewerApp: String;
    FHelpFilesOpened: TStringlist;
    FUrlSyntax: String;
    FDefWindowPos: TPoint;
    FDefWindowSize: TPoint;
    FDefStayOnTop: Boolean;
    procedure ValidateHelpViewer;
  public
    constructor Create;
    destructor Destroy; override;
    function  GetHelpFile(const Name: string): string;
    procedure InternalShutDown;
    { ICustomHelpViewer }
    function  GetViewerName: string;
    function  UnderstandsKeyword(const HelpString: string): Integer;
    function  GetHelpStrings(const HelpString: string): TStringList;
    function  CanShowTableOfContents: Boolean;
    procedure ShowTableOfContents;
    procedure ShowHelp(const HelpString: string);
    procedure NotifyID(const ViewerID: Integer);
    procedure SoftShutDown;
    procedure ShutDown;
    function  IsShowingHelp(const HelpFileName: string = ''): Boolean;
    { IExtendedHelpViewer }
    function  UnderstandsTopic(const Topic: string): Boolean;
    procedure DisplayTopic(const Topic: string); overload;
    function  UnderstandsContext(const ContextID: {$IFDEF CONTEXID_IS_THELPCONTEXT}THelpContext{$ELSE}Integer{$ENDIF}; const HelpFileName: string): Boolean;
    procedure DisplayHelpByContext(const ContextId: {$IFDEF CONTEXID_IS_THELPCONTEXT}THelpContext{$ELSE}Integer{$ENDIF}; const HelpFileName: string);
    { Execute the eWriter viewer app with the help string }
    function  CallEwriterHelp(const HelpFileName, HelpString: String; HelpContext: THelpContext; HelpClose: Boolean): Boolean;
    { Legacy methods linked to eWriter viewer }
    function  CallHtmlHelp(Handle: HWND; HelpFileName: string; Command: Integer; Data: NativeUInt): HWND;
    function  CallWinHelp(Handle: THandle; const HelpFileName: string; Command: Word; Data: NativeUInt): Boolean;
    { IExtendedHelpViewer2 }
    procedure DisplayTopic(const Topic, Anchor: string); overload;
    { properties }
    property ViewerID: Integer read FViewerID;
    property HelpManager: IHelpManager read FHelpManager write FHelpManager;
    property ViewerApp: String read FViewerApp write FViewerApp;
    property UrlSyntax: String read FUrlSyntax write FUrlSyntax;
    property DefWindowPos: TPoint read FDefWindowPos write FDefWindowPos;
    property DefWindowSize: TPoint read FDefWindowSize write FDefWindowSize;
    property DefStayOnTop: Boolean read FDefStayOnTop write FDefStayOnTop;
  end;

var
  HelpViewer: TEWriterHelpViewer;
  HelpViewerIntf: ICustomHelpViewer;

implementation

constructor TEWriterHelpViewer.Create;
begin
  inherited Create;
  HelpViewerIntf := Self;
  FInitialized := False;
  FViewerApp := 'eViewer.exe';           //This is a dummy name and needs to be replaced with a fully qualified path
  FUrlSyntax := 'index.html?%s.html%s';  //Standard url syntax for older Webhelp, also works with single-frame Webhelp
  //FUrlSyntax := '%s.html%s';             //Standard url syntax for new single-frame Webhelp

  FDefWindowPos := point(0,0);
  FDefWindowSize := point(0,0);
  FDefStayOnTop := false;

  FHelpFilesOpened := TStringlist.Create;
  FHelpFilesOpened.Sorted := true;
  FHelpFilesOpened.Duplicates := dupIgnore;
end;

destructor TEWriterHelpViewer.Destroy;
begin
  FHelpFilesOpened.Free;
  HelpViewer := nil;
  inherited Destroy;
end;

procedure TEWriterHelpViewer.ValidateHelpViewer;
resourcestring
  eEwriterViewerAppNotDefined = 'Unable to locate help viewer "%s"';
begin
  if not FInitialized then
  begin
    if FileExists(FViewerApp) then
      FInitialized := True
    else
      raise EHelpSystemException.Create(Format(eEwriterViewerAppNotDefined, [FViewerApp]));
  end;
end;

{ Tests if an instance of the viewer is running with this particular help file.
  Prevents help window from flashing when the app closes. }

function TEWriterHelpViewer.IsShowingHelp(const HelpFileName: string = ''): Boolean;
const
  MAX_PARAM_SIZE = 300;
  MAX_PARAMS = 10;
type
  PInstInfo = ^TInstInfo;
  TInstInfo = packed record
    FirstInstanceWnd: HWND;
    ParamCount: Integer;
    Params: Array[0..MAX_PARAMS-1, 0..MAX_PARAM_SIZE] of Char;
  end;
var
  HelpFile: string;
  lpInfo: PInstInfo;
  aSA: TSecurityAttributes;
  aSD: TSecurityDescriptor;
  MapFileName: String;
  MappingHandle: THandle;

  SemaphoreName: String;
begin
  result := false;
  HelpFile := GetHelpFile(HelpFileName);

  if HelpFile <> '' then
  begin
    { The eViewer creates a map file, based on the viewer app and the help file }
    MapFileName := ExtractFileName(FViewerApp) {$IFDEF EVIEWER250}+ ExtractFileName(HelpFileName) + 'Sem'{$ENDIF};

    InitializeSecurityDescriptor(@aSD, SECURITY_DESCRIPTOR_REVISION);
    SetSecurityDescriptorDacl(@aSD, True, nil, False);
    aSA.nLength := SizeOf(TSecurityAttributes);
    aSA.bInheritHandle := true;
    aSa.lpSecurityDescriptor := @aSD;

    MappingHandle := CreateFileMapping(INVALID_HANDLE_VALUE, @aSA, PAGE_EXECUTE_READWRITE, 0, sizeof(TInstInfo), PChar(MapFileName));
    if MappingHandle <> 0 then
    begin
      result := (GetLastError = ERROR_ALREADY_EXISTS);
      CloseHandle(MappingHandle);
    end;
  end;
end;


{ ICustomHelpViewer. }

function TEWriterHelpViewer.GetViewerName: string;
begin
  Result := 'EWriterHelpViewer';
end;

function TEWriterHelpViewer.UnderstandsKeyword(const HelpString: string): Integer;
begin
  Result := 1;
end;

function TEWriterHelpViewer.GetHelpStrings(const HelpString: string): TStringList;
begin
  Result := TStringList.Create;
end;

function TEWriterHelpViewer.CanShowTableOfContents: Boolean;
begin
  Result := True;
end;

procedure TEWriterHelpViewer.ShowTableOfContents;
var
  FileName : string;
begin
  ValidateHelpViewer;
  CallEwriterHelp('', '', 0, false);
end;

procedure TEWriterHelpViewer.ShowHelp(const HelpString: string);
begin
  DisplayTopic(HelpString);
end;

procedure TEWriterHelpViewer.NotifyId(const ViewerId: Integer);
begin
  FViewerID := ViewerID;
end;

procedure TEWriterHelpViewer.SoftShutDown;
var
  intIndex: Integer;
begin
  if FInitialized then
    for intIndex := FHelpFilesOpened.Count-1 downto 0 do
      if IsShowingHelp(FHelpFilesOpened[intIndex]) then
        CallEwriterHelp(FHelpFilesOpened[intIndex], '', 0, true);
end;

procedure TEWriterHelpViewer.ShutDown;
begin
  SoftShutDown;
  FInitialized := false;
  if Assigned(FHelpManager) then
    HelpManager := nil;
end;

{ IExtendedHelpViewer }

function TEWriterHelpViewer.UnderstandsTopic(const Topic: string): Boolean;
var
  HelpFile: string;
begin;
  HelpFile := GetHelpFile('');
  Result := HelpFile <> '';
end;

procedure TEWriterHelpViewer.DisplayTopic(const Topic: string);
var
  tmpTopic, tmpAnchor: string;
begin
  tmpTopic := Topic;
  tmpAnchor := '';
  if pos('#', tmpTopic) > 0 then
  begin
    tmpAnchor := copy(tmpTopic, pos('#', tmpTopic));
    tmpTopic := copy(tmpTopic, 1, pos('#', tmpTopic)-1);
  end;
  DisplayTopic(tmpTopic, tmpAnchor);
end;

function TEWriterHelpViewer.UnderstandsContext(const ContextId: {$IFDEF CONTEXID_IS_THELPCONTEXT}THelpContext{$ELSE}Integer{$ENDIF}; const HelpFileName: string): Boolean;
var
  HelpFile: string;
begin
  HelpFile := GetHelpFile('');
  Result := HelpFile <> '';
end;

procedure TEWriterHelpViewer.DisplayHelpByContext(const ContextId: {$IFDEF CONTEXID_IS_THELPCONTEXT}THelpContext{$ELSE}Integer{$ENDIF}; const HelpFileName: string);
begin
  CallEwriterHelp(HelpFileName, '', ContextID, false);
end;

function TEWriterHelpViewer.CallEwriterHelp(const HelpFileName, HelpString: String; HelpContext: THelpContext; HelpClose: Boolean): Boolean;
var
  FileName: string;
  defpos: String;
  cmd: String;
begin
  ValidateHelpViewer;
  Result := false;
  FileName := GetHelpFile(HelpFileName);
  if FileName <> '' then
  begin
    defpos := '';
    if fDefWindowPos <> point(0,0) then
    begin
      defpos := Format(' /defpos:%d,%d', [DefWindowPos.X, DefWindowPos.Y]);
      if fDefWindowSize <> point(0,0) then
      begin
        defpos := defpos + Format(',%d,%d', [DefWindowSize.X, DefWindowSize.Y]);
        if fDefStayOnTop then
          defpos := defpos + ',stayontop';
      end;
    end;

    if HelpClose then
      cmd := Format('"%s" close', [Filename])
    else if HelpContext <> 0 then
      cmd := Format('"%s" %d%s', [Filename, HelpContext, defpos])
    else begin
      cmd := Format('"%s"', [Filename]);
      if HelpString <> '' then
        cmd := cmd + Format(' "%s"', [HelpString]);
      if defpos <> '' then
        cmd := cmd + defpos;
    end;

    ShellExecute(0, 'open', Pchar(ViewerApp), PChar(cmd), '', SW_SHOWNORMAL);
    if HelpClose and (FHelpFilesOpened.Indexof(Filename) > -1) then
      FHelpFilesOpened.Delete(FHelpFilesOpened.Indexof(Filename))
    else
      FHelpFilesOpened.Add(Filename);
    Result := true;
  end;
end;

function TEWriterHelpViewer.CallWinHelp(Handle: THandle; const HelpFileName: string; Command: Word; Data: NativeUInt): Boolean;
begin
  result := false;
  case Command of
    HELP_CONTEXT: result := CallEwriterHelp(HelpFileName, '', Data, false);
    HELP_QUIT:    result := CallEwriterHelp(HelpFileName, '', 0, true);
    HELP_INDEX,
    HELP_HELPONHELP,
    HELP_FINDER:  result := CallEwriterHelp(HelpFileName, '', 0, false);
  else
    Exit;
  end;
end;

{ ISpecialHtmlHelpViewer }

function TEWriterHelpViewer.CallHtmlHelp(Handle: HWND; HelpFileName: string; Command: Integer; Data: NativeUInt): HWND;
begin
  Result := 0;
  case Command of
    HH_CLOSE_ALL:
    begin
      SoftShutDown;
    end;
    HH_HELP_CONTEXT:
    begin
      DisplayHelpByContext(Data, HelpFileName);
    end;
    else
      ShowTableOfContents;
  end;
end;


procedure TEWriterHelpViewer.DisplayTopic(const Topic, Anchor: string);
var
  tmpAnchor: String;
begin
  { Anchor can be blank, but if defined, a leading # is required }
  tmpAnchor := Anchor;
  if (tmpAnchor <> '') then
    if (tmpAnchor[1] <> '#') then
      tmpAnchor := '#' + tmpAnchor;

  CallEwriterHelp('', Format(UrlSyntax, [Topic, tmpAnchor]), 0, false);
end;

function TEWriterHelpViewer.GetHelpFile(const Name: string): string;
var
  FileName: string;
begin
  Result := '';
  if (Name = '') and Assigned(FHelpManager) then
    FileName := HelpManager.GetHelpFile
  else
    FileName := Name;

//  if (FileName = '') and Assigned(HtmlHelpTester) then
//    FileName := HtmlHelpTester.GetDefaultHelpFile;
//
//  if Assigned(HtmlHelpTester) then
//    FileName := HtmlHelpTester.GetHelpPath + PathDelim + FileName;

  Result := FileName;
end;

procedure TEWriterHelpViewer.InternalShutDown;
begin
  SoftShutDown;
  if Assigned(FHelpManager) then
  begin
    HelpManager.Release(ViewerID);
    HelpManager := nil;
  end;
end;

initialization
  HelpViewer := TEWriterHelpViewer.Create;
  System.HelpIntfs.RegisterViewer(HelpViewerIntf, HelpViewer.FHelpManager);

finalization
  if Assigned(HelpViewer.HelpManager) then
    HelpViewer.InternalShutDown;
  HelpViewerIntf := nil;

end.
